/*
        flc2gc.c -- generic container 2.0
        FlexList wrapper.

        (C) Copyright 1996  John Webster Small
        All rights reserved
*/

#include "flc2gc.h"

#ifndef FL_NDEBUG
void FlexList_delete(FlexList * L)
{ (void) FLdelete(&L); }
gc_delete_def(FlexList);
static gc * _FlexLists = 0;
void FL_trace_on()
{
    if (!_FlexLists)  {
        _FlexLists = gc_init_type(NEW_GC,FlexList);
        gc_resetFlags(_FlexLists,GC_ALL_FLAGS);
        gc_setFlags(_FlexLists,GC_BIND_ONLY);
    }
}
FILE * FL_stderr = 0;
void FL_trace(FILE * fp)
{
    FL_stderr = fp;
}
void FL_trace_report()
{
    FlexL L;
    if (_FlexLists)  {
        if (!FL_stderr)
            FL_stderr = stderr;
        gc_forEach(_FlexLists)  {
            L = THIS_FL(gc_get(_FlexLists));
            fprintf(FL_stderr,"%s FlexList: %p  "
              "nodes: %u  LINE: %d  FILE: %s\n",
              ((THIS_GC(L)->dynamic)? "Dynamic" : "Static "),
              L,FLnodes(L),L->LINE,L->FILENAME);
        }
    }
}
void FL_trace_off_report()
{
    FL_trace_report();
    if (_FlexLists)
        _FlexLists = gc_destruct(_FlexLists);
}
void FL_add_trace(FlexList * L)
{
    if (_FlexLists)
        (void) gc_insQ(_FlexLists,L);
}
void FL_rmv_trace(FlexList * L)
{
    if (_FlexLists)
        (void) gc_atRmv(_FlexLists,
            gc_indexOf(_FlexLists,L));
}
#endif

void * _FLnewNode(FlexL L)
{
    if (FLsizeofNodeData(L))
        return malloc(FLsizeofNodeData(L));
    return 0;
}

gc_vf_assignD(FlexNode)
/* void *  FlexNode_assignD(const gc * this_gc, void * D,
 const void * S, unsigned flags) */
{
    if (THIS_FL(this_gc)->sizeofNodeData)
        return memcpy(D,S,THIS_FL(this_gc)->sizeofNodeData);
    else if (flags & GC_READ)  {
        if (THIS_FL(this_gc)->vft &&
            THIS_FL(this_gc)->vft->FNread &&
            (*THIS_FL(this_gc)->vft->FNread)(S,D))
            return D;
    }
    else if (flags & GC_READ_DEL)  {
        if (THIS_FL(this_gc)->vft &&
            THIS_FL(this_gc)->vft->FNdestruct &&
            (*THIS_FL(this_gc)->vft->FNdestruct)((void *)S,D))
            return D;
    }
    else if (THIS_FL(this_gc)->vft &&
            THIS_FL(this_gc)->vft->FNwrite &&
            (*THIS_FL(this_gc)->vft->FNwrite)(D,S))
        return D;
    return 0;
}

gc_vf_newD(FlexNode)
/* void *  FlexNode_newD(const gc * this_gc, const void * D) */
{
    if (THIS_FL(this_gc)->sizeofNodeData)  {
        void * N = (void *)
            GC_NEW_DATA(THIS_FL(this_gc)->sizeofNodeData);
        if (N)
            memcpy(N,D,THIS_FL(this_gc)->sizeofNodeData);
        return N;
    }
    else if (THIS_FL(this_gc)->vft &&
        THIS_FL(this_gc)->vft->FNnew)
        return (void *) (*THIS_FL(this_gc)->vft->FNnew)(D);
    return 0;
}

gc_vf_deleteD(FlexNode)
/* void FlexNode_deleteD(const gc * this_gc, TYPE * D) */
{
    if (THIS_FL(this_gc)->FNdelete)
        (*(THIS_FL(this_gc)->FNdelete))(D);
    else
        GC_DELETE_DATA(D);
}

gc_vft_def(FlexNode,
    (gc_vF_assignD) (gc_vf_name(FlexNode,assignD)),
    (gc_vF_newD) (gc_vf_name(FlexNode,newD)),
    (gc_vF_deleteD) (gc_vf_name(FlexNode,deleteD)),
    (gc_vF_attachD) (0),
    (gc_vF_detachD) (0));

#ifdef FL_NDEBUG
FlexL FLinit(FlexL L_0, size_t sizeofNodeData, void (*FNdelete)(void *),
    FlexNVFT vft, size_t sizeofLocalData, int (*FLDdestruct)(void *))
#else
FlexL FLinit(const char * FILENAME, int LINE,
    FlexL L_0, size_t sizeofNodeData, void (*FNdelete)(void *),
    FlexNVFT vft, size_t sizeofLocalData, int (*FLDdestruct)(void *))
#endif
{
    unsigned short dynamic = 0;
    if (!L_0)  {
        dynamic = 1;
        L_0 = FL_NEW_HEADER(sizeofLocalData);
        if (!L_0)
            return FlexL0;
    }
    THIS_GC(L_0)->dynamic = dynamic;
#ifndef FL_NDEBUG
    FL_add_trace(L_0);
    assert(FILENAME);
    strncpy(L_0->FILENAME,FILENAME,SIZEOF_FlexList_FILENAME);
    L_0->FILENAME[SIZEOF_FlexList_FILENAME-1] = '\0';
    L_0->LINE = LINE;
#endif
    if (!sizeofNodeData && !vft)
        L_0 = THIS_FL(gc_init_type(THIS_GC(L_0),_gc_string));
    else
        L_0 = THIS_FL(gc_init_type(THIS_GC(L_0),FlexNode));
    THIS_GC(L_0)->dynamic = dynamic;
    L_0->sizeofNodeData = sizeofNodeData;
    L_0->FNdelete = FNdelete;
    L_0->vft = vft;
    L_0->FLDdestruct = FLDdestruct;
    return L_0;
}

#ifndef FL_NDEBUG
void FLdestruct(FlexL L)
{
    FL_rmv_trace(L);
    FLclear(L);
    assert(!THIS_GC(L)->dynamic);
}
#endif

int FLdelete(FlexL * Lptr)
{
    int ok = 1;
    assert(Lptr);
    assert(THIS_GC(*Lptr)->dynamic);
#ifndef FL_NDEBUG
    FL_rmv_trace(*Lptr);
#endif
    if ((*Lptr)->FLDdestruct)
        if (!(*((*Lptr)->FLDdestruct))((*Lptr)->data))
            ok = 0;
    if (!FLclear(*Lptr))
        ok = 0;
    *Lptr = THIS_FL(gc_destruct(THIS_GC(*Lptr)));
    return ok;
}

int FLstoreD(FlexL L, const void *D, unsigned n)
{
    if (n) (void) gc_setCurNode(THIS_GC(L),n-1);
    return gc_putAsg(THIS_GC(L),D) != 0;
}

int FLrecallD(FlexL L, void *D, unsigned n)
{
    if (n) (void) gc_setCurNode(THIS_GC(L),n-1);
    return gc_getAsg(THIS_GC(L),D) != 0;
}
