/*
        flc2gc.h -- generic container 2.0
        FlexList wrapper.

        (C) Copyright 1996  John Webster Small
        All rights reserved
*/

#ifdef  FLEXLIST_CPP
#error  Include "flc2gc.h" instead of "flexlist.hpp"
#endif

#ifndef flc2gc_h
#define flc2gc_h

/* LINTLIBRARY */

#include "gc.h"

/*
#define FL_NDEBUG
*/

#define FL_NOT_SUPPORTED assert(0)

#ifdef NDEBUG
#define FL_NDEBUG
#endif

#define FlexNodeLinkage
typedef struct {
    FlexNodeLinkage
    char data[1];
} FlexNode;
typedef FlexNode *FlexN;
#define FlexN0  ((FlexN)0)

typedef struct {
    FlexN (*FNnew)(const void *D);
    int   (*FNwrite)(void *ND, const void *D);
    int   (*FNread)(const void *ND, void *D);
    int   (*FNdestruct)(void *ND, void * D);
} FlexNodeVFT, * FlexNVFT;

#define FlexNVFT0  ((FlexNVFT)0)
#define FNnew0 ((FlexN (*)(const void *)) 0)
#define FNwrite0 ((int (*)(void *, const void *)) 0)
#define FNread0 ((int (*)(const void *,void *)) 0)
#define FNdestruct0 ((int (*)(void *, void *)) 0)

#ifndef FL_NDEBUG
#define SIZEOF_FlexList_FILENAME 21
#endif

typedef struct {
    struct gc_header gc_base;
    size_t sizeofNodeData;
    void (*FNdelete)(void *);
    FlexNVFT vft;
    int (*FLDdestruct)(void *);
#ifndef FL_NDEBUG
    char FILENAME[SIZEOF_FlexList_FILENAME];
    int LINE;
#endif
    char data[1];
} FlexList, * FlexL;
#define THIS_GC(L) ((gc *)L)
#define THIS_FL(GC) ((FlexL)GC)
#define FlexL0  ((FlexL)0)
#define FLcomparE(compare)  ((voidCmP) compare)
#define FLcompare0  FLcomparE(0)
#define FNdeleteCast(FNC) ((void (*)(void *))FNC)
#define FNdelete0 FNdeleteCast(0)
#define FLDdestructCast(FNC) ((int (*)(void *)) FNC)
#define FLDdestruct0 FLDdestructCast(0)
#define FL_NEW_HEADER(sizeofLocalData) \
    ((FlexL)malloc(sizeof(FlexList)-1+sizeofLocalData))

#ifndef FL_NDEBUG
extern void FlexList_delete(FlexList * L);
gc_decl(FlexList);
#define gc_FlexList(this_fl) \
    gc_init_type(THIS_GC(this_fl),FlexList)
#include <stdio.h>
extern void FL_trace(FILE * fp);
extern void FL_trace_on(void);
extern void FL_trace_report(void);
extern void FL_trace_off_report(void);
#define FL_TRACE(FP) FL_trace(FP);
#define FL_TRACE_ON  FL_trace_on();
#define FL_TRACE_REPORT  FL_trace_report();
#define FL_TRACE_OFF_REPORT FL_trace_off_report();
#else
#define FL_TRACE(FP)
#define FL_TRACE_ON
#define FL_TRACE_REPORT
#define FL_TRACE_OFF_REPORT
#endif

/* For internal use only! */
extern void * _FLnewNode(FlexL L);
#define FLD_2_GC(L,primitive,D)  \
    ( D ? gc_ ## primitive ## New(THIS_GC(L), D ) \
    : gc_ ## primitive (THIS_GC(L),_FLnewNode(L)))

extern gc_vf_assignD(FlexNode);
extern gc_vf_newD(FlexNode);
extern gc_vf_deleteD(FlexNode);
gc_decl(FlexNode);

/* FlexList contructors: */

#ifdef FL_NDEBUG
extern FlexL FLinit
(
    FlexL L_0,
    size_t sizeofNodeData,
    void (*FNdelete)(void *),
    FlexNVFT vft,
    size_t sizeofLocalData,
    int (*FLDdestruct)(void *)
);
#define FLfixed(L,sizeofNodeData) \
    FLinit(L,sizeofNodeData,FNdelete0,FlexNVFT0,0,FLDdestruct0)
#define FLtype(L,TYPE) \
    FLinit(L,sizeof(TYPE),FNdeleteCast(TYPE ## _delete), \
      FlexNVFT0,0,FLDdestruct0)
#define FLunpack(L,sizeofCell,cells,array) FL_NOT_SUPPORTED
#define FLvariant(L,vft) FLinit(L,0,FNdelete0,vft,0,FLDestruct0)
#define FLstr(L) FLvariant(L,0)

#define FLfixedNew(sizeofNodeData,sizeofLocalData,FLDdestruct) \
    FLinit(0,sizeofNodeData,FNdelete0,FlexNVFT0,sizeofLocalData,FLDdestruct)
#define FLtypeNew(TYPE,sizeofLocalData,FLDdestruct) \
    FLinit(0,sizeof(TYPE),FNdeleteCast(TYPE ## _delete), \
      FlexNVFT0,sizeofLocalData,FLDdestruct)
#define FLunpackNew(sizeofCell,cells,array,sizeofLocalData,FLDdestruct) \
    FL_NOT_SUPPORTED
#define FLvariantNew(vft,sizeofLocalData,FLDdestruct) \
    FLinit(0,0,FNdelete0,vft,sizeofLocalData,FLDdestruct)
#define FLstrNew(sizeofLocalData,FLDdestruct) \
    FLvariantNew(0,sizeofLocalData,FLDdestruct)
#else
extern FlexL FLinit
(
    const char * FILENAME,
    int LINE,
    FlexL L_0,
    size_t sizeofNodeData,
    void (*FNdelete)(void *),
    FlexNVFT vft,
    size_t sizeofLocalData,
    int (*FLDdestruct)(void *)
);
#define FLfixed(L,sizeofNodeData) \
    FLinit(__FILE__,__LINE__,L,sizeofNodeData, \
        FNdelete0,FlexNVFT0,0,FLDdestruct0)
#define FLtype(L,TYPE) \
    FLinit(__FILE__,__LINE__,L,sizeof(TYPE), \
        FNdeleteCast(TYPE ## _delete), \
        FlexNVFT0,0,FLDdestruct0)
#define FLunpack(L,sizeofCell,cells,array) FL_NOT_SUPPORTED
#define FLvariant(L,vft) \
    FLinit(__FILE__,__LINE__,L,0,FNdelete0,vft,0,FLDestruct0)
#define FLstr(L) FLvariant(L,0)

#define FLfixedNew(sizeofNodeData,sizeofLocalData,FLDdestruct) \
    FLinit(__FILE__,__LINE__,0,sizeofNodeData, \
        FNdelete0,FlexNVFT0,sizeofLocalData,FLDdestruct)
#define FLtypeNew(TYPE,sizeofLocalData,FLDdestruct) \
    FLinit(__FILE__,__LINE__,0,sizeof(TYPE), \
        FNdeleteCast(TYPE ## _delete), \
        FlexNVFT0,sizeofLocalData,FLDdestruct)
#define FLunpackNew(sizeofCell,cells,array,sizeofLocalData,FLDdestruct) \
    FL_NOT_SUPPORTED
#define FLvariantNew(vft,sizeofLocalData,FLDdestruct) \
    FLinit(__FILE__,__LINE__,0,0,FNdelete0,vft,sizeofLocalData,FLDdestruct)
#define FLstrNew(sizeofLocalData,FLDdestruct) \
    FLvariantNew(0,sizeofLocalData,FLDdestruct)
#endif

/* FlexList destructors: */

#define FLclear(L) gc_allDel(THIS_GC(L))
#ifdef FL_NDEBUG
#define FLdestruct(L) FLclear(L)
#else
extern void FLdestruct(FlexL L);
#endif
extern int FLdelete(FlexL * Lptr);

/* FlexList header functions: */

#define FLfrontD(L) gc_top(THIS_GC(L))
#define FLcurrentD(L) gc_get(THIS_GC(L))
#define FLrearD(L) gc_rear(THIS_GC(L))
#define FLcurNum(L) (gc_curNode(THIS_GC(L))+1)
#define FLnodes(L) gc_nodes(THIS_GC(L))
#define FLmaxNodes(L) gc_maxNodes(THIS_GC(L))
#define FLsetMaxNodes(L,maxNodes) gc_setMaxNodes(THIS_GC(L),maxNodes)
#define FLnotFull(L) gc_vacancy(THIS_GC(L))
#define FLsizeofNodeData(L) (L? L->sizeofNodeData : 0)
#define FLisSorted(L) gc_sorted(THIS_GC(L))
#define FLunSort(L)  gc_unSort(THIS_GC(L))
#define FLcompare(L) gc_cmP(THIS_GC(L))
#define FLsetCompare(L,compare) gc_setCmP(THIS_GC(L),compare)
#define FLisFixed(L) FLsizeofNodeData(L)
#define FLisVariant(L) !FLsizeofNodeData(L)
#define FLData(L) (L? L->data : 0)

/* FlexList stack and queue functions: */

#define FLpushN(L,N)  gc_push(THIS_GC(L),N)
#define FLpushD(L,D) FLD_2_GC(L,push,D)
#define FLpopN(L)  ((FlexN) gc_pop(THIS_GC(L)))
#define FLpopD(L,D) \
  (D? ((int) gc_popDelAsg(THIS_GC(L),D)) : gc_popDel(THIS_GC(L)))
#define FLtopD(L,D) (D? gc_topAsg(THIS_GC(L),D) : gc_top(THIS_GC(L)))
#define FLinsQN(L,N) gc_insQ(THIS_GC(L),N)
#define FLinsQD(L,D) FLD_2_GC(L,insQ,D)

/* FlexList list functions: */

#define FLmkcur(L,n) gc_setCurNode(THIS_GC(L),(n?n-1:GC_NOTFOUND))
#define FLinsN(L,N) gc_insAfter(THIS_GC(L),N)
#define FLinsD(L,D) FLD_2_GC(L,insAfter,D)
#define FLinsSortN(L,N) gc_insSort(THIS_GC(L),N)
#define FLinsSortD(L,D) FLD_2_GC(L,insSort,D)
#define FLdelN(L) gc_rmvBackup(THIS_GC(L))
#define FLdelD(L,D) \
  (D? ((int) gc_delAsgBackup(THIS_GC(L),D)) : gc_delBackup(THIS_GC(L)))
#define FLnextD(L,D) (D? gc_nextAsg(THIS_GC(L),D) : gc_next(THIS_GC(L)))
#define FLprevD(L,D) (D? gc_prevAsg(THIS_GC(L),D) : gc_prev(THIS_GC(L)))

/* FlexList search/sort functions: */

#define FLfindFirstD(L,D) gc_findFirst(THIS_GC(L),D)
#define FLfindNextD(L,D) gc_findNext(THIS_GC(L),D)
#define FLfindLastD(L,D) FL_NOT_SUPPORTED
#define FLfindPrevD(L,D) FL_NOT_SUPPORTED
#define FLsort(L,compare) gc_sort_specs(THIS_GC(L),compare)

/* FlexList array functions: */

extern int FLstoreD(FlexL L, const void *D, unsigned n);
extern int FLrecallD(FlexL L, void *D, unsigned n);

/* FlexList compaction functions: */

#define FLpack(L) FL_NOT_SUPPORTED
#define FLpackPtrs(L) FL_NOT_SUPPORTED

#define FLmaxMaxNodes  GC_MAXNODES
#define FLnewAlignLoss  16
#define FLmaxSizeofNodeData  \
    ((size_t)(-(long)sizeof(FlexNode) \
    -FLnewAlignLoss))
#define FLmaxSizeofArray ((long)(size_t)-FLnewAlignLoss)

#endif
